
#include <qstring.h>
//#include <qvaluevector.h>
#include "levenshtein.h"

/*
 * This implementation is based on the one found here:
 * http://www.merriampark.com/ldcpp.htm
 * 
 * This file is part of Dixit.
 * Author: Octavian Procopiuc <oprocopiuc@gmail.com>
 * Created: July 25, 2004
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You can get the GNU General Public License from
 * http://www.gnu.org/licenses/licenses.html
 */

// If maxdist == 0, return the true Levenshtein (edit) distance.
// Otherwise, return the minimum between maxdist and the edit distance.
// If source or target are longer than 50, return -1.
int levenshtein(const QString& source, const QString& target, int maxdist) {

  char n, m;
  uint l;
  l = source.length();
  if (l > 50) return -1;
  n = l;
  
  l = target.length();
  if (l > 50) return -1;
  m = l;

  if (maxdist == 0) maxdist = QMAX(m, n);
  if (n == 0) return QMIN(m, maxdist);
  if (m == 0) return QMIN(n, maxdist);

  // Store the min. value on each column, so that, if it reaches
  // maxdist, we break early.
  char mincolval;

  char matrix[51][51];

  char j;
  char i;
  char cell;

  for (j = 0; j <= m; j++)
    matrix[0][j] = j;

  for (i = 1; i <= n; i++) {

    mincolval = QMAX(m, i);
    matrix[i][0] = i;

    const QChar s_i = source.at(i-1);

    for (j = 1; j <= m; j++) {

      const QChar t_j = target.at(j-1);

      char cost = (s_i == t_j ? 0 : 1);

      const char above = matrix[i-1][j];
      const char left = matrix[i][j-1];
      const char diag = matrix[i-1][j-1];
      cell = QMIN( above + 1, QMIN(left + 1, diag + cost) );

      // Cover transposition, in addition to deletion,
      // insertion and substitution. This step is taken from:
      // Berghel, Hal ; Roach, David : "An Extension of Ukkonen's 
      // Enhanced Dynamic Programming ASM Algorithm"
      // (http://www.acm.org/~hlb/publications/asm/asm.html)
      
      if (i > 2 && j > 2) {
        char trans = matrix[i-2][j-2] + 1;
        if (source.at(i-2) != t_j) trans++;
        if (s_i != target.at(j-2)) trans++;
        if (cell > trans) cell = trans;
      }
      
      mincolval = QMIN(mincolval, cell);
      matrix[i][j] = cell;
    }

    if (mincolval >= maxdist)
      break;

  }

  if (i == n + 1)
    return (int) matrix[n][m];
  else
    return maxdist;
}
