/*
 * This file is part of Dixit.
 * Author: Octavian Procopiuc <oprocopiuc@gmail.com>
 * Created: July 12, 2004
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You can get the GNU General Public License from
 * http://www.gnu.org/licenses/licenses.html
 */

/****************************************************************************
** ui.h extension file, included from the uic-generated form implementation.
*****************************************************************************/

#include "dex_abrev.h"
#include "dixit_strings.h"
 
const uint MAX_DEFINITIONS_DISPLAYED = 200;
const uint MAX_HISTORY_SIZE = 58;
const QString DEFAULT_SERVER_NAME = "dexonline.ro";
const uint DEFAULT_SERVER_PORT = 80;
const QString HTTP_GET_REQUEST = "/update.php?timestamp=%1&compress=gzip";
const QString WINDOWS_REGISTRY = "/Dictionaries";
const QString APP_KEY = "/Dixit/";
//const time_t LAST_UPDATE_TIME_BASE = 1090000000; // For debugging.
const time_t LAST_UPDATE_TIME_BASE = 0;
// Don't change this key! It will break backwards compatibility.
const char * LAST_UPDATE_TIME_KEY = "<>";
const QString BROWSER_DBTOTAL = 
  QString::fromAscii("Baza de date con\\tine <b>%1</b> %2defini\\tii din diverse <u><a href=\"!surse\">surse</a></u>.");



// Index the database in a different thread.
class buildAllTitlesThread: public QThread {
  mainForm* form;
 public:
  buildAllTitlesThread(mainForm *f): QThread(), form(f) { }
  virtual void run() {
    form->buildAllTitles();
  }
};



void mainForm::makeFontBigger()
{
  QFont cf = font();
  int newsz = cf.pointSize() + 1;
  if (newsz > 16) newsz = 16;
  cf.setPointSize(newsz);
  setFont(cf);
}

void mainForm::makeFontSmaller()
{
  QFont cf = font();
  int newsz = cf.pointSize() - 1;
  if (newsz < 8) newsz = 8;
  cf.setPointSize(newsz);
  setFont(cf);
}

void mainForm::print()
{
  // Taken from QT's examples.
#ifndef QT_NO_PRINTER
    QPrinter printer(QPrinter::PrinterResolution);
    printer.setFullPage(TRUE);
    printer.setPageSize(printerPageSize);
    if ( printer.setup( this ) ) {
        QPainter p( &printer );
        QPaintDeviceMetrics metrics(p.device());
        int dpix = metrics.logicalDpiX();
        int dpiy = metrics.logicalDpiY();
        const int margin = 54; // pt
        QRect body(margin*dpix/72, margin*dpiy/72,
                   metrics.width()-margin*dpix/72*2,
                   metrics.height()-margin*dpiy/72*2 );
        QSimpleRichText richText( browser->text(), QFont(), browser->context(), browser->styleSheet(),
                                  browser->mimeSourceFactory(), body.height() );
        richText.setWidth( &p, body.width() );
        QRect view( body );
        int page = 1;
        do {
            richText.draw( &p, body.left(), body.top(), view, colorGroup() );
            view.moveBy( 0, body.height() );
            p.translate( 0 , -body.height() );
            p.drawText( view.right() - p.fontMetrics().width( QString::number(page) ),
                        view.bottom() + p.fontMetrics().ascent() + 5, QString::number(page) );
            if ( view.top()  >= richText.height() )
                break;
            printer.newPage();
            page++;
        } while (TRUE);
	printerPageSize = printer.pageSize();
    }
#endif
}

void mainForm::changeFont()
{
  bool ok;
  QFont nf = QFontDialog::getFont(&ok, font(), this );
  if ( ok ) {
    if (nf.pointSize() > 16)
      nf.setPointSize(16);
    else if (nf.pointSize() < 8)
      nf.setPointSize(8);
    setFont(nf);
  }
}


void mainForm::fileExit()
{
  QApplication::exit(0);
}


void mainForm::helpAbout()
{
  linkFind(QString::fromAscii("!info"));
}

void mainForm::setDefaultSource(int id)
{
  defaultSource = id;
  sourceBox->changeItem(0, sourceMenu->text(id));
  sourceBox->update(); // For windows.
}

void mainForm::setStatusFromSource(int id)
{
  QString tmpstr = DIXIT_SOURCES[id][2];
  diacriticizeString(tmpstr);
  statusBar()->message(tmpstr);
}


void mainForm::lineFind()
{
    QString word = searchLine->text();
    if (word.isEmpty()) return;

    // Handle certain commands right away.
    if (word[0] == '!') {
      if (word.lower() == "!maimare") {
	makeFontBigger();
	searchLine->setText(currentWord);
	return;
      } else if (word.lower() == "!maimic") {
	makeFontSmaller();
	searchLine->setText(currentWord);
	return;
      } 
    }

    if (currentWord.length() > 0) {
	if (backwardWordList.count() >= MAX_HISTORY_SIZE)
	    backwardWordList.pop_front();
	backwardWordList.push_back(currentWord);
	forwardWordList.clear();
    }
 
    insertInHistory(word);
 }


void mainForm::historyFind()
{
    QString word = historyList->currentText();    
    if (word.isEmpty() || (word == currentWord)) return;
    if (currentWord.length() > 0) {
	if (backwardWordList.count() >= MAX_HISTORY_SIZE)
	    backwardWordList.pop_front();
	backwardWordList.push_back(currentWord);
	forwardWordList.clear();
    }
    insertInLine(word);
    insertInHistory(word);    
}

void mainForm::linkFind( const QString & word )
{
     if (word.isEmpty()) return;

     // Handle certain commands right away.
     if (word[0] == '!') {
       if (word.lower() == "!maimare") {
	 makeFontBigger();
	 searchLine->setText(currentWord);
	 return;
       } else if (word.lower() == "!maimic") {
	 makeFontSmaller();
	 searchLine->setText(currentWord);
	 return;
       } 
     }

     if (currentWord.length() > 0) {
	 if (backwardWordList.count() >= MAX_HISTORY_SIZE)
	     backwardWordList.pop_front();
	 backwardWordList.push_back(currentWord);
	 forwardWordList.clear();
     }
     
     QString lword(word.lower());
     if (lword[0] != '!')
       lword.replace(QRegExp("[^a-z*?|#\\- ]"), "");
     
     insertInLine(lword);
     insertInHistory(lword);
}

void mainForm::insertInHistory( const QString & word )
{
    // Limit the history.
    if (historyList->count() >= MAX_HISTORY_SIZE)
	historyList->removeItem(historyList->count() - 1);
    
    // Insert word at top of history list.
    QListBoxItem * item;
    if ((item = historyList->findItem(word, Qt::ExactMatch))) {
	historyList->setSelected(item, true);
    } else {
	historyList->insertItem(word, 0);
	historyList->setCurrentItem(0);
    }
    genericFind(word);
    currentWord = word;
    setFocusToLine();
}

void mainForm::backwardInHistory()
{
    if (backwardWordList.isEmpty()) return;
    if (historyList->selectedItem()) forwardWordList.push_back(historyList->currentText());
    insertInLine(backwardWordList.last());
    insertInHistory(backwardWordList.last());
    backwardWordList.pop_back();
}


void mainForm::forwardInHistory()
{
    if (forwardWordList.isEmpty()) return;
    if (historyList->selectedItem()) backwardWordList.push_back(historyList->currentText());
    insertInLine(forwardWordList.last());
    insertInHistory(forwardWordList.last());
    forwardWordList.pop_back();
}


void mainForm::insertInLine( const QString & word )
{
    // Insert word on the edit line.
    searchLine->setText(word);
}


void mainForm::genericFind( const QString & word)
{
    QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );

    if (pthread)
      pthread->wait();

    QStringList wordlist;
    QStringList::iterator sit;
    QValueList<QCString>::iterator it;
    QString tmpbs;
    QString buf;
    QRegExp rx;
    QMap<QString,QString> defmap;

    uint exactmatches; // number of exact word matches.
    QString lword = word.lower();
    datum key, data;

    // Ignore spaces.
    lword.replace(QRegExp("[ ]"), "");

    if (lword.startsWith("dixit") || lword.startsWith("ipse"))
      lword = "!dixit";

    // Check for commands.
    if (lword[0] == '!') {

      QString lanchor = lword.section('#', 1);
      lword = lword.section('#', 0, 0);
      clearBrowser();
      outToBrowser(commandOutput(lword));
      if (!lanchor.isEmpty()) {
	browser->scrollToAnchor(lanchor);
      }
      searchLine->selectAll();
      goto genericFind_done;
    }

    // Ignore dashes.
    lword.replace(QRegExp("[\\-]"), "");
    
    if (lword.isEmpty()) {
      goto genericFind_done;
    }

    clearBrowser();

    if ( !dbOpenRead() ) {
      tmpbs = QString::fromAscii(
	    "<font color='#993300'>Eroare: Nu pot citi din baza de date.</font><br>"
	    "Verifica\\ti c~a baza de date exist~a (%1). Dac~a exist~a, s-ar putea s~a "
	    "fie ^in curs de actualizare de c~atre alt~a instan\\t~a a programului, "
	    "sau s-ar putea s~a fie corupt~a.<br>")
        .arg(dbFileName);

      diacriticizeString(tmpbs);
      outToBrowser(tmpbs);
      dbClose();
      goto genericFind_done;
    }
    

    // Find all matching words by searching in allTitles.

    buf = QString::fromAscii("^(") + lword;
    buf.replace('*', "[a-z]*");
    buf.replace('?', "[a-z]");
    buf += QString::fromAscii(")$");
    // Case sensitive, since everything should be lower case.
    rx = QRegExp(buf, true);
    rx.setMinimal(true);

    for (it = allTitles.begin();  it != allTitles.end(); it++) {
      QString sbuf = (*it).data();
      
      if (sbuf.find(rx) != -1) {

      	wordlist.append(sbuf);

	// Check for other definitions for this title.
        // They are stored in the database with keys title10, title11,...
        int duplicates;
        int ij = 10;
        QCString awij;
        do {
	  awij.sprintf("%s%d", (*it).data(), ij);
	  key.dptr = awij.data();
	  key.dsize = awij.length();
	  if ((duplicates = gdbm_exists(dbFile, key))) {
	    wordlist.append(QString::fromAscii(awij.data()));
	  }
	  ij++;
        } while (duplicates);

      }
    }

    exactmatches = wordlist.size();
    
    // If no exact matches, try to find close matches.
    if ( !exactmatches && !lword.contains('*') && !lword.contains('?') && !lword.contains('|')) {
      int edist;

      for (it = allTitles.begin();  it != allTitles.end(); it++) {

	QString sbuf = (*it).data();
	if (lword.length() + 2u <= sbuf.length() || lword.length() >= sbuf.length() + 2u)
	  continue;

	edist = levenshtein(lword, sbuf, 2);
	if (edist < 2) {

	  wordlist.append(sbuf);

	  // Check for other definitions for this title.
	  // They are stored in the database with keys title10, title11,...
	  int duplicates;
	  int ij = 10;
	  QCString awij;
	  do {
	    awij.sprintf("%s%d", (*it).data(), ij);
	    key.dptr = awij.data();
	    key.dsize = awij.length();
	    if ((duplicates = gdbm_exists(dbFile, key))) {
	      wordlist.append(QString::fromAscii(awij.data()));
            }
            ij++;
	  } while (duplicates);

	  if (wordlist.size() >= MAX_DEFINITIONS_DISPLAYED / 2)
	    break;
	}
      }
    }

  
    // Fetch the definitions and store them in a map.
    if (wordlist.size()) {

	QString buffer;
	QCString lbuf;
	QString realtitle; // Spaced out title, suitable for comparison in defmap.
	sit = wordlist.begin();
	QCString srcstr(">"); // The end of the <S> (source) tag.
	srcstr += DIXIT_SOURCES[defaultSource][0];

	while (sit != wordlist.end() && defmap.size() <= MAX_DEFINITIONS_DISPLAYED) {
	    QCString aw = (*sit).ascii();
	    key.dptr = aw.data();
	    key.dsize = aw.length();
	    
	    data = gdbm_fetch (dbFile, key);

	    //buffer = "";
	    if (data.dptr != NULL) {
	      lbuf.fill('\0', data.dsize + 1);
	      strncpy(lbuf.data(), data.dptr, data.dsize);
	      free(data.dptr);
	      // Find the real title from the definition.
	      realtitle = findRealTitle(*sit, lbuf);

	      if (lbuf.find(srcstr.data()) == -1) {
		sit++;
		continue;
	      }
	      buffer = QString::fromUtf8(processXMLEntry(lbuf));
	    } else {
	      buffer = QString::fromAscii(
                "<p><font color='#993300'>Eroare intern~a: "
	  	"nu am g~asit cuv^antul %1 ^in baza de date.</font></p>")
		.arg(*sit);

	      diacriticizeString(buffer);
	      qWarning("*** Warning: word in list not found in db: %s.", (*sit).ascii());
	      realtitle = findRealTitle(*sit, "");
	    }

	    defmap[realtitle] = buffer;
	    sit++;
	}

    }

    // tmpbs collects the string that will be sent to the browser.
    tmpbs = "<table cellpadding=\"0\" cellspacing=\"1\" width=\"100%\"><tr><td bgcolor=\"#eeeeee\">";

    if (exactmatches) {

      QCString lbuf;
      if (defaultSource != 0 && defmap.size() < exactmatches) {
	if (defmap.size() <= MAX_DEFINITIONS_DISPLAYED)
	  lbuf.sprintf(" din diverse surse (%d din %s)",
		       defmap.size(), DIXIT_SOURCES[defaultSource][1]);
	else
	  lbuf.sprintf(" din diverse surse (mai mult de %d din %s)",
		       MAX_DEFINITIONS_DISPLAYED, DIXIT_SOURCES[defaultSource][1]);
      }

      // Weird case that appears with duplicate entries.
      if (defaultSource == 0 && defmap.size() < exactmatches && defmap.size() <= MAX_DEFINITIONS_DISPLAYED)
	exactmatches = defmap.size();

      tmpbs += QString::fromAscii("C~autarea \"%3\" a returnat %2 %4defini\\ti%1%5.</td></tr>")
	.arg(exactmatches == 1 ? "e" : "i").arg(exactmatches)
	.arg(word).arg((exactmatches % 100) >= 20 ? "de " : "")
	.arg(QString::fromUtf8(lbuf.data()));

    } else {

      tmpbs += QString::fromAscii("C~autarea \"%1\" nu a returnat nici o defini\\tie.").arg(word);

    }

    tmpbs += "<tr><td align=\"right\"><font size=-1 color='#686868'>"
             "<a href=\"!abr\">Abrevieri</a></font></td></tr>";
    tmpbs += "</table>";
    
    if (!exactmatches && defmap.size()) {
      tmpbs += QString::fromAscii(
          "<table cellpadding=\"0\" cellspacing=\"1\" width=\"100%\">"
	  "<tr><td><font color=\"#993300\">Cele mai apropiate rezultate sunt:</font></td></tr></table>\n");
    }

    diacriticizeString(tmpbs);

    if (defmap.size() <= MAX_DEFINITIONS_DISPLAYED) {

      QMap<QString,QString>::Iterator mit;

      // Concatenate all definitions (already diacriticized and htmlized).
      for ( mit = defmap.begin(); mit != defmap.end(); ++mit ) {
	tmpbs += mit.data();
      }

    } else {
      QString tmptmp = QString::fromAscii(
		       "<p><font color=\"#993300\">Num~arul maxim de defini\\tii care pot fi "
			"afi\\sate este %1. <br>Folosi\\ti un criteriu de c~autare mai "
			"selectiv.</font></p>")
	.arg(MAX_DEFINITIONS_DISPLAYED);

      diacriticizeString(tmptmp);
      tmpbs += tmptmp;
    }
    

    browser->setText(tmpbs);
    //    outToBrowser(tmpbs);

    dbClose();

genericFind_done:
    QApplication::restoreOverrideCursor();

}


void mainForm::init()
{
    pthread = new buildAllTitlesThread(this);

    // Load settings (window geometry, font size, default source, printer page size).
    loadSettings();
    statusBar()->setFont(toolBar->font());
    QToolTip::setFont(toolBar->font());

    setIcon( QPixmap::fromMimeSource( "dixit_books22.png" ) );

    int iarg = 1;
    // Jump over potential command line options.
    while (qApp->argc() > iarg && qApp->argv()[iarg][0] == '-') {
      iarg++;
    }

    QDir dbd; // Current direcory.
    dbd.convertToAbs();

    // Set up the database file name.
    if (qApp->argc() > iarg) {
      strncpy(dbFileName, QDir::cleanDirPath(dbd.filePath(qApp->argv()[iarg])).ascii(), 128);
    } else {
      strncpy(dbFileName, dbd.filePath(DIXIT_DB_FILE_NAME), 128);
    }

    qDebug("*** Database file: %s", dbFileName);

    // The last update time is stored in the database under this key.
    // (cannot make it const due to gdbm function restrictions)
    strcpy( lastUpdateTime_key, LAST_UPDATE_TIME_KEY );

    // Clear the right panel.
    historyList->clear();

    http = new QHttp(this);

    connect( http, SIGNAL(responseHeaderReceived(const QHttpResponseHeader &)),
	     this, SLOT(testResponseHeader(const QHttpResponseHeader &)) );
    connect( http, SIGNAL(readyRead(const QHttpResponseHeader &)),
	     this, SLOT(processRawData(const QHttpResponseHeader &)) );
    connect( http, SIGNAL(done(bool)), this, SLOT(closeConnection(bool)) );

    connect( &connectionTimer, SIGNAL(timeout()),
		this, SLOT(connectionTimeout()) );
    //disconnect( browser, SIGNAL(linkClicked(const QString &)), 0, 0);
    //browser->setMimeSourceFactory(0);
    connect( browser, SIGNAL(linkClicked(const QString &)), 
	     this, SLOT(linkFind(const QString &)));
    connect( browser, SIGNAL(highlighted(const QString &)), 
	     (const QObject*) statusBar(), SLOT(message(const QString &)));


    // Find last update time from the database.
    datum key, data;
    if ( !dbOpenRead() ) {
        // Probably a new installation.
	lastUpdateTime = LAST_UPDATE_TIME_BASE;
	qWarning("*** I will create a new database.");
    } else {
	key.dptr = lastUpdateTime_key;
	key.dsize = strlen(lastUpdateTime_key);
	data = gdbm_fetch(dbFile, key);
	if (!data.dptr) {
	    qWarning("*** Internal error: No timestamp found in the database.");
	    lastUpdateTime = LAST_UPDATE_TIME_BASE;
	} else {
	    QCString luts;
	    appendToCString(luts, data.dptr, data.dsize);
	    free(data.dptr);
	    lastUpdateTime = luts.toLong();
	}
    }

    dbClose();
    

    // Do not underline hyperlinks.
    browser->setLinkUnderline(false);
    browser->setTextFormat(RichText);

    searchLine->setMaxLength(50);

    // Allow only the following on the search line: letters, *, ?, |, -, space.
    QRegExp re("!{0,1}[a-zA-Z*?|#\\- ]+");
    searchLine->setValidator(new QRegExpValidator(re, this));
    searchLine->setFrameShadow(QFrame::Sunken);
    
    sourceBox = new QMenuBar(toolBar, "sourceBox");
    sourceMenu = new QPopupMenu(this);
    QFont tbf = toolBar->font();
    sourceMenu->setFont(tbf); // Not bold.
    tbf.setWeight(QFont::Bold);
    sourceBox->setFont(tbf); // Bold.
    sourceBox->setPaletteForegroundColor( QColor( 4, 102, 186 ) );

    sourceMenu->insertItem(QString::fromUtf8(DIXIT_SOURCES[0][1]), 0);
    int ij;
    for (ij = 6; ij < DIXIT_SOURCES_COUNT; ij++)
      sourceMenu->insertItem(QString::fromUtf8(DIXIT_SOURCES[ij][1]), ij);

    // defaultSource is set in loadSettings.
    sourceBox->insertItem(QString::fromUtf8(DIXIT_SOURCES[defaultSource][1]), sourceMenu, 0);
    sourceBox->setFrameShape(QFrame::NoFrame);
    //sourceBox->setMinimumWidth(sourceBox->fontMetrics().width(sourceBox->text(0)) + 10);        

    connect(sourceMenu, SIGNAL(activated(int)), this, SLOT(setDefaultSource(int)) );
    connect(sourceMenu, SIGNAL(highlighted(int)), this, SLOT(setStatusFromSource(int)) );

    QToolTip::add(sourceBox, 
		  QString::fromUtf8("Filtreaz\304\203 rezultatele afi\305\237ate dup\304\203 surs\304\203"));

    // I moved these here from the ui-generated file since I cannot insert sourceMenu from the ui.
    editFindAction->addTo( toolBar );
    toolBar->addSeparator();
    updateAction->addTo( toolBar );
    aboutAction->addTo( toolBar );
    fileExitAction->addTo( toolBar );


    // This does not seem to have any effect, paragraphs are still left-aligned.
    browser->setAlignment(Qt::AlignJustify);

    // Set link color to dark blue, etc.
    QPalette pal = browser->palette();
    pal.setColor(QPalette::Active,   QColorGroup::Link, QColor("#003399"));
    pal.setColor(QPalette::Active,   QColorGroup::Text, QColor("#191919"));
    pal.setColor(QPalette::Inactive, QColorGroup::Link, QColor("#003399"));
    pal.setColor(QPalette::Inactive, QColorGroup::Text, QColor("#191919"));
    pal.setColor(QPalette::Disabled, QColorGroup::Link, QColor("#3366cc"));
    pal.setColor(QPalette::Disabled, QColorGroup::Text, QColor("#555555"));
    browser->setPalette(pal);

    clearBrowser();

    if (autoUpdate || gdbm_errno == GDBM_FILE_OPEN_ERROR) {
      // Send a special info command to the browser.
      outToBrowser(commandOutput("!infointro"));
      // Start a connection to the server to check for updates.
      genericUpdate();
    } else {
      // Build an in-memory word index, asynchronously.
      pthread->start();
      // Send a special info command to the browser.
      outToBrowser(commandOutput("!infointro"));
    }

    // Set keyboard focus to the search line.
    setFocusToLine();
 }


void mainForm::saveSettings()
{
    QSettings settings;
    settings.insertSearchPath( QSettings::Windows, WINDOWS_REGISTRY );
    settings.writeEntry( APP_KEY + "WindowWidth", width() );
    settings.writeEntry( APP_KEY + "WindowHeight", height() );
    settings.writeEntry( APP_KEY + "WindowX", x() );
    settings.writeEntry( APP_KEY + "WindowY", y() );
    settings.writeEntry( APP_KEY + "AutoUpdate", autoUpdate );
    settings.writeEntry( APP_KEY + "PrinterPageSize", (int) printerPageSize );
    settings.writeEntry( APP_KEY + "DefaultSource", defaultSource );
    settings.writeEntry( APP_KEY + "FontPointSize", font().pointSize() );
    settings.writeEntry( APP_KEY + "FontFamily", font().family() );
}


void mainForm::loadSettings()
{
    QSettings settings;
    settings.insertSearchPath( QSettings::Windows, WINDOWS_REGISTRY );
    int windowWidth = settings.readNumEntry( APP_KEY + "WindowWidth", 770 );
    int windowHeight = settings.readNumEntry( APP_KEY + "WindowHeight", 500 );
    int windowX = settings.readNumEntry( APP_KEY + "WindowX", 100 );
    int windowY = settings.readNumEntry( APP_KEY + "WindowY", 100 );
    autoUpdate = settings.readBoolEntry( APP_KEY + "AutoUpdate", true );
    printerPageSize = 
      (QPrinter::PageSize) settings.readNumEntry( APP_KEY + "PrinterPageSize", (int) QPrinter::A4 );
    defaultSource = settings.readNumEntry( APP_KEY + "DefaultSource", 0 );
#if defined(Q_OS_WIN)
    QString fontFamily = "Arial";
#else
    QString fontFamily = "Sans";
#endif
    fontFamily = settings.readEntry( APP_KEY + "FontFamily", fontFamily );
    int fontPointSize = settings.readNumEntry( APP_KEY + "FontPointSize", 10 );
    resize( windowWidth, windowHeight );
    move( windowX, windowY );
    QFont cf = font();
    if (fontPointSize < 8) fontPointSize = 8;
    if (fontPointSize > 16) fontPointSize = 16;
    cf.setPointSize(fontPointSize);
    cf.setFamily(fontFamily);
    cf.setStyleStrategy(QFont::PreferAntialias);
    setFont(cf);
}

// Connect to the server and get updates.
// This slot is linked to the update button.
void mainForm::buttonUpdate()
{
    searchLine->clear();
    historyList->clearSelection();
    clearBrowser();
    if (currentWord.length() > 0) {
	if (backwardWordList.count() >= MAX_HISTORY_SIZE)
	    backwardWordList.pop_front();
	backwardWordList.push_back(currentWord);
	forwardWordList.clear();
	currentWord = "";
    }

    genericUpdate();
}


// Start a connection with the server.
// Don't link this slot directly to a signal, 
// make a wrapper (like buttonUpdate).
void mainForm::genericUpdate()
{    
    pthread->wait();

    QString lbuf = QString::fromAscii(
          "<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\"><tr><td bgcolor=\"#eeeeee\">"
	  "<b>Actualizare</b></td></tr><tr><td>^Incerc s~a contactez serverul %1... ")
      .arg(DEFAULT_SERVER_NAME);
    diacriticizeString(lbuf);
    outToBrowser(lbuf);
    searchLine->clearFocus();
    toolBar->setEnabled(false);
    browser->setEnabled(false);
    historyList->setEnabled(false);

    updateError = UPDATE_ERR_NO_ERROR;
    lastUpdateTime_pending = lastUpdateTime;

    QString tmpstr;
    tmpstr = HTTP_GET_REQUEST;
    QHttpRequestHeader header( "GET", tmpstr.arg(lastUpdateTime) );
    header.setValue("Host", DEFAULT_SERVER_NAME);
    tmpstr = DIXIT_USER_AGENT;
    header.setValue( "User-Agent", tmpstr.arg(DIXIT_VERSION) );
    header.setValue( "Accept-Encoding", "gzip" );
    header.setValue( "Connection", "Keep-Alive" );
    // Timeout after 20 seconds.
    connectionTimer.start(20 * 1000, true);
    connectionStarting = true;
    newDefCount = 0;
    newDefCount_pending = -1;

    http->setHost(DEFAULT_SERVER_NAME);
    http->request(header);
}

// Slot linked to connectionTimer.
void mainForm::connectionAbort() {
  if (http->state() != QHttp::Unconnected && http->state() != QHttp::Closing) {
    updateError = UPDATE_ERR_ABORTED;
    http->abort();
  }
}

// Slot linked to the connectionTimer.
void mainForm::connectionTimeout() {
  updateError = UPDATE_ERR_TIMEOUT;
  http->abort();
}

// Slot linked to QHttp::done.
void mainForm::closeConnection(bool error)
{
  QString wc;
  connectionTimer.stop();

  // Close the database.
  dbClose();

  if (!error && newDefCount_pending == -1) {
    updateError = UPDATE_ERR_CONTENT;
    error = true;
  }

  if (error) {
    // Treat some special cases.
    if (updateError == UPDATE_ERR_TIMEOUT || updateError == UPDATE_ERR_ABORTED) 
      // "Minor" error. Recover gracefully below.
      error = false;
    else if (!updateError) 
      // We never got a chance to set updateError, probably due to a network error.
      updateError = UPDATE_ERR_NETWORK;
  }


  if (updateError != UPDATE_ERR_DB_READ && updateError != UPDATE_ERR_DB_WRITE) {
    lastUpdateTime = lastUpdateTime_pending;

    if ( dbOpenWrite() ) {
      datum key, data;
      QCString luts;
      luts.sprintf("%d", lastUpdateTime);
      // Store the new update time in the database.
      key.dptr = lastUpdateTime_key;
      key.dsize = strlen(lastUpdateTime_key);
      data.dptr = luts.data();
      data.dsize = luts.length();
      
      gdbm_store(dbFile, key, data, GDBM_REPLACE);
      dbClose();
    }

  }

  if (error) {
    qWarning("*** Error during update: %s", UPDATE_ERRORS_STRING[updateError]);
    QString wc = QString::fromAscii(
        "<font color='#993300'>Eroare (%1).</font><br> "
	"Data ultimei actualiz~ari (sau a ultimului cuv^ant actualizat): %2.<br>")
      .arg(UPDATE_ERRORS_STRING[updateError]).arg(time2str(lastUpdateTime));

    if (updateError == UPDATE_ERR_DB_READ || updateError == UPDATE_ERR_DB_WRITE)
      wc += QString::fromAscii("Eroare ^in baza de date: <i>%1</i><br>")
	.arg(gdbm_strerror((gdbm_error) gdbm_errno));

    diacriticizeString(wc);
    outToBrowser(wc);

  } else {


    wc = QString::fromAscii(" Terminat. (%1 %2defini\\tii noi sau actualizate).<br>")
      .arg(newDefCount).arg(newDefCount % 100 >= 20 ? "de " : "");
    if (newDefCount) {
      wc += QString::fromAscii("Cele mai noi defini\\tii pot fi g~asite <u><a href=\"!nou\">aici</a></u> (maxim 20).<br>");
    }
    if (historyList->count() == 0 && autoUpdate) {
      wc += QString::fromAscii(
	  "<br>Baza de date se actualizeaz~a automat la pornirea programului. "
	  "Pentru a schimba aceast~a preferin\\t~a ap~asa\\ti <u><a href=\"!lazimanual\">aici</a></u>. "
	  " Actualizarea se poate face ^in orice moment folosind butonul de actualizare.<br>");
    }
    diacriticizeString(wc);
    outToBrowser(wc);

    if (updateError == UPDATE_ERR_TIMEOUT || 
	updateError == UPDATE_ERR_NOT_DONE || 
	updateError == UPDATE_ERR_ABORTED) {
      qWarning("*** Warning: Connection timeout or abort. Download cut short.");
      wc = "<font color=\"#993300\">Leg~atura s-a ^intrerupt prematur. "
	   "Ap~asa\\ti butonul de actualizare pentru a desc~arca restul de defini\\tii.</font><br>";
      qDebug("Last Update Time: %d", (int) lastUpdateTime);
      diacriticizeString(wc);
      outToBrowser(wc);   
    } else {
      qDebug("*** Finished all data from server.");
    }
  }

  // Build an in-memory word list.
  pthread->start();

  browser->setEnabled(true);
  historyList->setEnabled(true);
  toolBar->setEnabled(true);
  setFocusToLine();
}

// This slot is linked to the QHttp::responseHeaderReceived signal.
void mainForm::testResponseHeader( const QHttpResponseHeader & rh ) {
  if (rh.statusCode() != 200) {
    qWarning("*** Warning: HTTP Server returned status code %d %s",
	     rh.statusCode(), rh.reasonPhrase().ascii());
    updateError = UPDATE_ERR_SERVER;
    http->abort();
  } else if (!(rh.contentType().contains("gzip") || rh.value(QString::fromAscii("content-encoding")).contains("gzip"))) {
    qWarning("*** Warning: Unknown content type received from server: %s", rh.contentType().ascii());
    updateError = UPDATE_ERR_CONTENT;
    http->abort();
  }
}

// This slot is linked to the QHttp::readyRead signal.
void mainForm::processRawData( const QHttpResponseHeader & rh )
{
    connectionTimer.start(20 * 1000, true);

    static QCString xmlData;
    static bool gzipped = false;
    
    if (connectionStarting) {
      // These are redundant...
      newDefCount = 0;
      newDefCount_pending = -1;
      // These are also redundant...
      if (rh.statusCode() != 200) {
	qWarning("*** Warning: HTTP Server returned status code %d %s",
		 rh.statusCode(), rh.reasonPhrase().ascii());
	updateError = UPDATE_ERR_SERVER;
	http->abort();
	return;
      }

      qDebug("*** HTTP Server returned status code %d %s",
		 rh.statusCode(), rh.reasonPhrase().ascii());

      // Initialize the zlib inflate structures, if necessary.
      if (rh.contentType().contains("gzip") || 
	  rh.value(QString::fromAscii("content-encoding")).contains("gzip")) {
	gzipped = true;
	processGZIPData(xmlData, NULL, 0, true);
      } else {
	qWarning("*** Warning: Unknown content type received from server: %s", rh.contentType().ascii());
	updateError = UPDATE_ERR_CONTENT;
	http->abort();
	return;
      }

      xmlData.truncate(0);
      connectionStarting = false;
    }

    uint datalen = xmlData.length();

    if (datalen < 100000) {
      //    while (true) {
      uint ba = http->bytesAvailable();
      //qDebug("\tbytes %d", ba);
      if ( !ba ) {
	//break;
	qWarning("*** Warning: processRawData called, but no data is available.");
	//	updateError = UPDATE_ERR_STREAM;
	//	http->abort();
	return;
      }

      // Don't read more than this at a time.      
      if (ba > 10000)
        ba = 10000;

      char * buf = new char[ba + 1];
      int buf_size = http->readBlock( buf, ba );
      if (buf_size == -1) {
	qWarning("*** Warning: Error occured during download (readBlock returned -1).");
	updateError = UPDATE_ERR_STREAM;
	http->abort();
	delete [] buf;
	return;
      }
      
      // Inflate, if needed.
      if (gzipped) {
	processGZIPData(xmlData, buf, buf_size, false);
      } else {
	appendToCString(xmlData, buf, buf_size);
      }
      delete [] buf;
      
    }

    bool alldone = false;
    datalen = xmlData.length();

    alldone = xmlData.contains("</Dictionary>");
    
    // Process xml data when 20k chars are available. There is nothing special about 20k here.
    if (datalen >= 20000 || alldone) {
      uint lastidx = processXMLData(xmlData);
      // If anything processed, eat it.
      if (lastidx != 0)
	xmlData = xmlData.mid(lastidx, datalen);
      QString wc = QString::fromAscii(
                   " <b>%1/%2</b> defini\\tii noi sau actualizate "
		   "(ap~asa\\ti <i>Esc</i> pentru a ^intrerupe).")
	.arg(newDefCount).arg(newDefCount_pending);
      diacriticizeString(wc);
      outToBrowser(wc, false);
    }

    if (http->bytesAvailable()) {
      emit moreData(rh);
    }
}



// Process XML from C-string data. Return the number of chars processed.
uint mainForm::processXMLData( const QCString & data )
{
    // This is the last update time advertised by the server at the
    // beginning of the transmition. We use it only when all data was
    // received ok. Otherwise, we use the individual times in each
    // entry.
    static time_t lut_advertised = 0;
    updateError = UPDATE_ERR_NOT_DONE;
    uint lastIdx = 0;
    uint idx = 0;
    int ij;
    // Indexes of the open tags.
    uint entryoIdx = 0, entrycIdx = 0, idoIdx = 0, timeoIdx = 0, nroIdx = 0;
    // Indexes of the closing tags.
    uint nameoIdx = 0, namecIdx = 0, idcIdx = 0, timecIdx = 0, nrcIdx = 0;
    QCString lbufname(100);
    QCString lbufnamedup(100);
    QCString lbufentry;
    QCString lbufentrydup(101);
    datum crtName, crtEntry, dupe;
    //qDebug("*** processXMLData: Start.");
    uint datalen = data.length();
    while (idx < datalen) {
        if (data.at(idx) != '<') { 
	  idx++; 
	  continue;
	}

	char nextch1 = data.at(idx+1);
	char nextch2 = data.at(idx+2);

	// Check if we haven't seen an entry yet.
	if (entryoIdx == 0) {
	// This is a bit deangerous, since it relies on NumResults appearing after Timestamp.
	//if (newDefCount_pending == -1) {
	  // Test for the update timestamp.
	  if (nextch1 == 'T' && data.mid(idx, 11) == "<Timestamp>") {
	    timeoIdx = idx + 11;
	    idx += 11;
	    continue;
	  } else if (nextch1 == '/' && nextch2 == 'T' && data.mid(idx, 12) == "</Timestamp>") {
	    timecIdx = idx;
	    lut_advertised = data.mid(timeoIdx, timecIdx - timeoIdx).toLong();
	    idx += 12;
	    continue;
	  } else if (nextch1 == 'N' && data.mid(idx, 12) == "<NumResults>") {
	    nroIdx = idx + 12;
	    idx += 12;
	    continue;
	  } else if (nextch1 == '/' && nextch2 == 'N' && data.mid(idx, 13) == "</NumResults>") {
	      nrcIdx = idx;
	      newDefCount_pending = data.mid(nroIdx, nrcIdx - nroIdx).toInt();
	      if ( newDefCount_pending ) {
		// Open the database.
		if ( !dbOpenWrite() ) {
		  updateError = UPDATE_ERR_DB_WRITE;
		  http->abort();
		  return 0;
		}
	      }
	      idx += 13;
	      continue;
	  }
	} 
	
	if (nextch1 == '/' && nextch2 == 'D' && data.mid(idx, 13) == "</Dictionary>") {
	  if (lut_advertised > 0)
	    lastUpdateTime_pending = lut_advertised;
	  dbClose();
	  updateError = UPDATE_ERR_NO_ERROR;
	  http->closeConnection();
	  idx += 13;
	} else if (nextch1 == 'T' && nextch2 == 'i') { //data.mid(idx, 11) == "<Timestamp>") {
	    timeoIdx = idx + 11;
	    idx += 11;
	} else if (nextch1 == '/' && nextch2 == 'T' && data.at(idx+3) == 'i') { //data.mid(idx, 12) == "</Timestamp>") {
	    timecIdx = idx;
	    idx += 12;
	} else if (nextch1 == 'N' && nextch2 == 'a') { //data.mid(idx, 6) == "<Name>") {
	    nameoIdx = idx + 6;
	    idx += 6;
	} else if (nextch1 == '/' && nextch2 == 'N' && data.at(idx+3) == 'a') { //data.mid(idx, 7) == "</Name>") {
	    namecIdx = idx;
	    idx += 7;
	} else if (nextch1 == 'I' && nextch2 == 'd') { //data.mid(idx, 4) == "<Id>") {
	    idoIdx = idx; // Include the <Id> tag!
	    idx += 4;
	} else if (nextch1 == '/' && nextch2 == 'I' && data.at(idx+3) == 'd') { //data.mid(idx, 5) == "</Id>") {
	    idcIdx = idx + 5; // Include the </Id> tag.
	    idx += 5;
	} else if (nextch1 == 'E' && nextch2 == 'n') {//data.mid(idx, 7) == "<Entry>") {
	    entryoIdx = idx + 7;
	    idx += 7;
	} else if (nextch1 == '/' && nextch2 == 'E' && data.at(idx+3) == 'n') { //data.mid(idx, 8) == "</Entry>") {

	    if (newDefCount_pending == -1) {
	      updateError = UPDATE_ERR_CONTENT;
	      http->abort();
	      return 0;
	    }
	    entrycIdx = idx;
	    lbufname = data.mid(nameoIdx, namecIdx - nameoIdx).lower().data();
	    if (entrycIdx - entryoIdx > 50000) {
		qWarning("*** Warning: Long definition: %d", entrycIdx - entryoIdx);
	    }
	    if (namecIdx - nameoIdx > 50) {
		qWarning("*** Warning: Long word: %d", namecIdx - nameoIdx);
	    }
	    if (!dbFile) {
	      qWarning("Internal Error: Database is not open during update.");
	      updateError = UPDATE_ERR_DB_WRITE;
	      http->abort();
	      return 0;
	    }
	    crtName.dptr = lbufname.data();
	    crtName.dsize = namecIdx - nameoIdx;
	    lbufentry.resize(entrycIdx - entryoIdx + 1);
	    ///lbufentry.fill('\0', entrycIdx - entryoIdx + 1);
	    strncpy(lbufentry.data(), &data.at(entryoIdx), entrycIdx - entryoIdx);
	    //crtEntry.dptr = &data.at(entryoIdx); ///llbuf.data();
	    //crtEntry.dsize = entrycIdx - entryoIdx; ///llbuf.size();
	    ij = 10;
	    //lbufentrydup.truncate(0);
	    while (gdbm_exists(dbFile, crtName)) {
		// Check whether this word exists with the same Id.
		dupe = gdbm_fetch(dbFile, crtName);
		if (dupe.dptr != NULL) {
		    // This is a bit dangerous, it relies on Id's being at the beginning of an entry.
		    lbufentrydup.fill('\0', 101);
		    strncpy(lbufentrydup.data(), dupe.dptr, QMIN(100, dupe.dsize)); // dupe.dsize);
		    free(dupe.dptr);
		    if (lbufentrydup.contains(data.mid(idoIdx, idcIdx - idoIdx).data()))
			break;

		} else {
		    qWarning("*** Internal error: word \"%s\" exists, but cannot be retrieved", lbufname.data());
		}
		
		lbufnamedup.sprintf("%s%d", lbufname.data(), ij);
		crtName.dptr = lbufnamedup.data();
		crtName.dsize = lbufnamedup.length();
		ij++;
	    }

	    if (gdbm_errno && gdbm_errno != GDBM_ITEM_NOT_FOUND) {
	      updateError = UPDATE_ERR_DB_READ;
	      http->abort();
	      return 0;
	    }
	    // Shorten the XML tags before storing them in the db.
	    lbufentry.replace("Name>", "N>");
	    lbufentry.replace("Definition>", "D>");
	    lbufentry.replace("Timestamp>", "T>");
	    lbufentry.replace("Author>", "A>");
	    lbufentry.replace("Source>", "S>");
	    lbufentry.replace("\n", "");
	    crtEntry.dptr = lbufentry.data();
	    crtEntry.dsize = lbufentry.length();
	    gdbm_store(dbFile, crtName, crtEntry, GDBM_REPLACE);
	    if (gdbm_errno && gdbm_errno != GDBM_ITEM_NOT_FOUND) {
	      updateError = UPDATE_ERR_DB_WRITE;
	      http->abort();
	      return 0;
	    }
	    lastUpdateTime_pending = data.mid(timeoIdx, timecIdx - timeoIdx).toLong() - 1;
	    newDefCount++;
	    if (newDefCount + 20 > newDefCount_pending) {
	      newDefList.push_front(lbufentry);
	      if (newDefList.size() > 20)
		newDefList.pop_back();
	    }
	    lastIdx = idx + 8;
	    idx += 8;
            
	} else {
	    idx++;
	}
    }

    //qDebug("*** processXMLData: End.");
    return lastIdx;
}


// Format an XML entry for the browser.
QCString mainForm::processXMLEntry( const QCString & data )
{
    uint i;
    uint idx = 0;
    uint oIdx[5];
    uint cIdx[5];
  
    // Open tags.
    static const QCString oTags[] = { "<N>", "<D>", "<S>", "<A>", "<T>" };
    // Close tags.
    static const QCString cTags[] = { "</N>", "</D>", "</S>", "</A>", "</T>" };

    uint datalen = data.length();
    while (idx < datalen ) {
	if (data.at(idx) != '<') { idx++; continue; }
	
	for (i = 0; i < 5; i++) {
	    if (data.mid(idx, oTags[i].length()) == oTags[i]) {
		oIdx[i] = idx + oTags[i].length();
	    } else if (data.mid(idx, cTags[i].length()) == cTags[i]) {
		cIdx[i] = idx;
	    }
	}
	idx++;
    }
    
    char sbuf[50000];
    htmlize( sbuf,  data.mid(oIdx[1], cIdx[1] - oIdx[1]) );
    QCString lbuf2 = "<table cellpadding=0 cellspacing=3 width=\"100%\"><tr><td>";
    lbuf2 += sbuf;
    lbuf2 += "<br><font size=-1 color='#686868'>Trimis de <b>";
    lbuf2 += data.mid(oIdx[3], cIdx[3] - oIdx[3]).data();
    lbuf2 += ".</b> Ultima actualizare: ";
    time_t tv = strtoul(data.mid(oIdx[4], cIdx[4] - oIdx[4]).data(), NULL, 10);
    lbuf2 += time2str(tv);
    lbuf2 += ". Sursa: ";
    lbuf2 += sourceString(data.mid(oIdx[2], cIdx[2] - oIdx[2]));
    lbuf2 += ".</font></td></tr></table>\n";
    return lbuf2;
}


void mainForm::outToBrowser( const QString & str, bool prs )
{
    browser->setText( browserBody + str ); 
    if (prs) { // persist.
	browserBody += str;
    }
}


QCString mainForm::time2str( time_t t, bool hour )
{
    if (t == 0)
      return QCString("Niciodat~a");

    time_t lt = t;
    tm ts; // time structure.
    QCString res(36);
    ts = *localtime(&lt);

    if (hour)
      res.sprintf("%d %s %d, %02d:%02d", ts.tm_mday, MONTH_NAMES[ts.tm_mon], ts.tm_year + 1900, ts.tm_hour, ts.tm_min);
    else
      res.sprintf("%d %s %d", ts.tm_mday, MONTH_NAMES[ts.tm_mon], ts.tm_year + 1900);

    return res;
}


// Store all titles from db in the allTitles list.
// Return the number of titles stored.
uint mainForm::buildAllTitles()
{

    if ( !dbOpenRead() ) {
        updateError = UPDATE_ERR_DB_READ;
	return 0;
    }

    datum key, nextkey;
    uint ans = 0;
    allTitles.clear();

    // Use the firstkey/nextkey mechanism in gdbm to iterate over all keys.
    key = gdbm_firstkey(dbFile);
    while (key.dptr) {
	
	if (key.dptr[0] != '<' && key.dptr[0] != '>') {
          ans++;

          // Store each word only once in the allTitles list, ie skip duplicates.
          // (duplicate titles have a number at the end in the database)	
	  if ( !(key.dptr[key.dsize-1] >= '0' && key.dptr[key.dsize-1] <= '9') ) {
	    QCString buf(key.dsize + 1);
	    strncpy(buf.data(), key.dptr, key.dsize);
	    buf.at(key.dsize) = '\0';
	    allTitles.append(buf);
	  }
	}

	nextkey = gdbm_nextkey(dbFile, key);
	free (key.dptr);
	key = nextkey;
    }

    dbClose();

    totalWordCount = ans;
    return ans;
}


void mainForm::setFocusToLine()
{
    searchLine->setFocus();
    searchLine->selectAll();
}


void mainForm::clearBrowser()
{
    browserBody = ""; 
    browser->clear();
}



// Process zdata and append inflated result to idata.
void mainForm::processGZIPData( QCString & idata, char* zdata, uint zdatasz, bool init_only )
{
    int err;
    uint len;
    bool done = false;
    static z_stream gz_stream;
    char gzbuffer[65536];
    if (init_only) {
	// Initialize zlib inflate.
	gz_stream.zalloc = NULL;
	gz_stream.zfree = NULL;
	gz_stream.opaque = NULL;
	gz_stream.next_in = gz_stream.next_out = 0;
	gz_stream.avail_in = gz_stream.avail_out = 0;
	if (inflateInit2(&gz_stream, 16+15) != Z_OK) {
	    qWarning("*** Error initializing gzip inflate. (zlib 1.2 needed)");
	}
	return;
    }
    
    gz_stream.next_out = (Bytef *) gzbuffer;
    gz_stream.avail_out = sizeof(gzbuffer);
    gz_stream.next_in = (Bytef *) (zdata);
    gz_stream.avail_in = zdatasz;
    while (gz_stream.avail_in != 0) {
	if (gz_stream.avail_out == 0) {
	    appendToCString(idata, gzbuffer, sizeof(gzbuffer));
	    gz_stream.next_out = (Bytef *) gzbuffer;
	    gz_stream.avail_out = sizeof(gzbuffer);
	}
	
	err = inflate(&gz_stream, Z_NO_FLUSH);
	if (err == Z_STREAM_END) {
	    break;
	} else if (err != Z_OK) {
	    qWarning("*** Error during inflate. Stream is not gzipped or zlib version is too old (zlib 1.2 needed).");
	    // This is a bit dramatic...
	    exit(1);
	}
    }

    while (true) {
	len = sizeof(gzbuffer) - gz_stream.avail_out;
	if (len > 0) {
	    appendToCString(idata, gzbuffer, len);
	    gz_stream.next_out = (Bytef *) gzbuffer;
	    gz_stream.avail_out = sizeof(gzbuffer);
	}
                                                                                                          
	if (done)
	    break;
	
	err = inflate(&gz_stream, Z_SYNC_FLUSH);
	if (len == 0 && err == Z_BUF_ERROR)
	    err = Z_OK;
	
	done = (gz_stream.avail_out != 0 || err == Z_STREAM_END);
	
	if (err != Z_OK && err != Z_STREAM_END)
	    break;
    }

}


void mainForm::appendToByteArray( QByteArray & ba, const char * buf, uint bufsz )
{
    uint oldsz = ba.size();
    ba.resize(oldsz + bufsz, QGArray::SpeedOptim);
    uint i;
    for (i = 0; i < bufsz; i++) {
    	ba.at(oldsz + i) = buf[i];
    }
}

void mainForm::appendToCString( QCString & cs, const char * buf, uint bufsz )
{
    uint oldsz = cs.length(); // NOT Including terminating '\0'
    cs.resize(oldsz + bufsz + 1);
    strncpy(&cs.at(oldsz), buf, bufsz);
    cs.at(oldsz + bufsz) = '\0';
}


QCString mainForm::sourceString( const QCString & src )
{
    QCString lsrc = src.lower();
    QCString ans;
    int ij;
    for (ij = 1; ij < DIXIT_SOURCES_COUNT; ij++) {
      if (lsrc == DIXIT_SOURCES[ij][0]) {
	ans.sprintf("<a href=\"!surse\">%s</a>", DIXIT_SOURCES[ij][1]);
	break;
      }
    }

    if (ij == DIXIT_SOURCES_COUNT)
      ans = src.upper();

    return ans;
}


QString mainForm::commandOutput( const QString & cmd )
{
    QString res;
    if (cmd.startsWith("!surs") || cmd.startsWith("!dex") || cmd == "!der" || cmd == "!dlra" || cmd == "!dor"
	|| cmd == "!dlrc" || cmd == "!dlrm" || cmd == "!dmlr" || cmd == "!doom" || cmd == "!mda"
	|| cmd == "!mdn" || cmd == "!neoficial" || cmd.startsWith("!anton")) {

      res = DIXIT_CMD_SOURCES;

    } else if (cmd.startsWith("!info") || cmd == "!despre") {
      res = QString::fromUtf8(
	  "<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\"><tr><td bgcolor=\"#eeeeee\">"
	  "<font size=\"+1\" color=\"#222222\"><b>"
	  "<a href=\"!dixit\">Dixit</a> - Dic\\tionar Explicativ Integrat"
	  "</b></font></td></tr><tr><td>Dixit permite c~aut~ari de cuvinte <i>offline,</i>"
	  " folosind baza de date a proiectului <i>DEX online.</i><br>") 
	+ (cmd == "!infointro" ? 
	   QString(" ") :
	   BROWSER_DBTOTAL.arg(totalWordCount).arg((totalWordCount) % 100 >= 20 ? "de " : "") 
	   + QString::fromUtf8(" Data ultimei actualiz~ari: %1.").arg(time2str(lastUpdateTime, true))
	  )
	+ QString::fromUtf8(
	  "</td></tr><tr><td bgcolor=\"#eeeeee\"><b>Instruc\\tiuni</b></td></tr>"
	  "<tr><td>Pentru a c~auta defini\\tia unui cuv^ant, tasta\\ti-l ^in cutia de deasupra \\si ap~asa\\ti <i>Enter.</i>"
	  " Pute\\ti c~auta mai multe defini\\tii simultan, folosind caracterele speciale \"<b>?</b>\" "
	  "(care corespunde oric~arei litere) \\si \"<b>*</b>\" "
	  " (care corespunde oric~arei secven\\te de litere, inclusiv secven\\ta vid~a). De exemplu, "
	  "c~autarea \"<u><a href=\"c?z*\">c?z*</a></u>\" va returna defini\\tiile "
	  "cuvintelor CAZ, CAZAN, CIZMAR etc. Pentru mai multe detalii, "
	  "vizita\\ti <u><a href=\"!instructiuni\">pagina de instruc\\tiuni</a></u>.</td></tr>"
	  "<tr><td bgcolor=\"#eeeeee\"><b>Licen\\t~a</b></td></tr>"
	  "<tr><td><p><b>Dixit</b> \302\251 2004 Octavian Procopiuc &lt;oprocopiuc@gmail.com&gt;. "
	  "Ave\\ti permisiunea de a copia, de a redistribui \\si de a modifica acest program "
	  "conform Licen\\tei Publice Generale GNU (GPL). <u><a href=\"!licenta\">Detalii</a></u>.<br><br>"
	  "<b>Lista de defini\\tii</b> \302\251 2004 <i>DEX Online</i> <b>www.dexonline.ro</b> "
	  "(coordonator: C~at~alin Fr^ancu &lt;cata@francu.com&gt;). "
	  "Ave\\ti permisiunea de a copia, de a redistribui \\si de a modifica defini\\tiile din "
	  "<i>DEX online</i> conform prevederilor din Licen\\ta GNU pentru Documenta\\tie Liber~a. "
	  "<u><a href=\"!licenta\">Detalii</a></u>.</td></tr></table>"
	  );
  } else if (cmd == "!licenta" || cmd == "!copiere") {

      res = DIXIT_CMD_LICENSE;

  } else if (cmd == "!instructiuni" || cmd == "!ajutor" || cmd == "!help") {

      res = QString::fromAscii(
	  "<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\">"
	  "<tr><td bgcolor=\"#eeeeee\"><b>Instruc\\tiuni</b></td></tr><tr><td>") 

	+ DIXIT_CMD_HELP_MAIN 

	+ QString::fromAscii(
	  "<br>Baza de date con\\tin^and defini\\tiile este p~astrat~a local \\si %1 se actualizeaz~a "
	  "la pornirea programului%2. "
	  "Pentru a schimba aceast~a preferin\\t~a ap~asa\\ti <u><a href=\"!lazi%3\">aici</a></u>. "
	  " Actualizarea se poate face ^in orice moment folosind butonul de actualizare.<br>")
	.arg(autoUpdate?"":"nu")
	.arg(autoUpdate?" dac~a exist~a o leg~atur~a la Internet":"")
	.arg(autoUpdate?"manual":"auto")

	+ DIXIT_CMD_HELP_KEYS

	+ QString::fromAscii("</td></tr></table>");

  } else if (cmd.startsWith("!laziauto") || cmd.startsWith("!lazimanu")) {

    res = QString::fromAscii("<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\">"
		 "<tr><td bgcolor=\"#eeeeee\"><b>Am schimbat modul de actualizare</b></td></tr><tr><td>");
    if (cmd.startsWith("!laziauto")) {
      autoUpdate = true;
    } else if (cmd.startsWith("!lazimanu")) {
      autoUpdate = false;
    }
    res += QString::fromAscii("Lista de defini\\tii <b>%1 se actualizeaz~a automat</b> la pornirea programului. "
		  "Pentru a schimba aceast~a preferin\\t~a ap~asa\\ti <u><a href=\"!lazi%2\">aici</a></u>."
		  "</td></tr></table>").arg(autoUpdate?"":"nu").arg(autoUpdate?"manual":"auto");

  } else if (cmd.startsWith("!dixit") || cmd.startsWith("!ipse")) {

    res = DIXIT_CMD_DIXIT;

  } else if (cmd.startsWith("!abr") || cmd.startsWith("!presc")) {

    res = DEX_ABREV;

  } else if (cmd == "!nou" || cmd == "!recent") {

    res = QString::fromAscii("<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\">"
		 "<tr><td bgcolor=\"#eeeeee\"><b>Cele mai noi defini\\tii</b></td></tr>");
    if (newDefList.isEmpty()) {
      res += QString::fromAscii("<tr><td><p>Nimic nou. Data ultimei actualiz~ari: %1.</p></td></tr></table>")
	.arg(time2str(lastUpdateTime));
    } else {
      res += QString::fromAscii("<tr><td align=\"right\"><font size=-1 color='#686868'>"
		    "<a href=\"!abrevieri\">Abrevieri</a></font></td></tr>");
      res += "</table>";
      
      QValueList<QCString>::iterator it;
      for ( it = newDefList.begin(); it != newDefList.end(); ++it ) {
	res += QString::fromUtf8(processXMLEntry(*it));
      }
    }

  } else {

    res = DIXIT_CMD_UNKNOWN;

  }

  diacriticizeString(res);
  return res;
}


// Given a title and its definition, return a title that's appropriate as key for the map used in the search.
// This is a bit tricky, since we need to have unique answers and the right order. Tread carefully.
QString mainForm::findRealTitle( const QString & title, const QCString & def )
{
    QString sdef(def);
    QRegExp realtitle_rx("<D>\\s*@([^@,]*)");
    QString tmp;
    QString res("");
    QString titlesuf = "0";
    QRegExp suf_rx("(\\d+)$");
    // If there are digits at the end of the title, add them in.
    if (suf_rx.search(title) != -1)
      titlesuf += suf_rx.cap(1);

    if (realtitle_rx.search(sdef) == -1) {
	qWarning("*** Warning: Could not find title in definition of \"%s\"", title.latin1());
	tmp = title;
    } else {
	tmp = realtitle_rx.cap(1);
	if (tmp.length() == 0)
	    tmp = title;
	else {
	  
	}
    }

    tmp = tmp.upper();
    tmp.replace(QRegExp("^A\\s+"), "");
    tmp.replace(QRegExp("^SE\\s+"), "");
    tmp.replace(QRegExp("^\\(SE\\)\\s+"), "");

    tmp += titlesuf;

    // Now space out tmp and place the result in res.
    uint i;
    QChar extrachar = ' ';
    for (i = 0; i < tmp.length(); i++) {
      if (tmp.at(i) == '\'' || tmp.at(i) == '`' || tmp.at(i) == ':' || tmp.at(i) == ' ' || tmp.at(i) == '-' || tmp.at(i) == '/') {
	    ; // Ignore silently;
      } else if (tmp.at(i) == '~') {
	extrachar = '-';
      } else if (tmp.at(i) == '\\' || tmp.at(i) == '^') {
	extrachar = tmp.at(i);
      } else if ((tmp.at(i) >= 'A' && tmp.at(i) <= 'Z')  || (tmp.at(i) >= '0' && tmp.at(i) <= '9')) {
	res += tmp.at(i);
	if (tmp.at(i) >= '1' && tmp.at(i) <= '9' && extrachar == '^')
	  res += QChar(' ');
	else
	  res += extrachar;
	extrachar = ' ';
      } else if (tmp.at(i) == '(' || tmp.at(i) == '&') {
	break;
      } else {
	qWarning("Warning: Unexpected character in computed title for \"%s\": %c", title.latin1(), tmp.at(i).latin1());
      }
    }

    return res;
}


// Primitive in-place parsing of diacritics directives a la DEX online.
// This is just to avoid Unicode in this file. 
// A true parsing is done in Catalin's htmlize.cpp
void mainForm::diacriticizeString( QString & s )
{
    uint i, slen = s.length();
    QChar extrachar = ' ';
    uint ni = 0; // the trailing iterator, for the result.
    for (i = 0; i < slen; i++) {
	if ((s.at(i) == '\\' || s.at(i) == '^' || s.at(i) == '~')) {
	    if (extrachar != ' ')
		s.at(ni++) = extrachar;
	    extrachar = s.at(i);
	    continue;
	}
	if (extrachar == '\\') {
	    if (s.at(i) == 'S') s.at(ni++) = 350;
	    else if (s.at(i) == 's') s.at(ni++) = 351;
	    else if (s.at(i) == 'T') s.at(ni++) = 354;
	    else if (s.at(i) == 't') s.at(ni++) = 355;
	    else { s.at(ni++) = '\\'; s.at(ni++) = s.at(i); }
	} else if (extrachar == '~') {
	   if (s.at(i) == 'A') s.at(ni++) = 0x102;
	   else if (s.at(i) == 'a') s.at(ni++) = 0x103;
	   else { s.at(ni++) = '~'; s.at(ni++) = s.at(i); }
	} else if (extrachar == '^') {
	    if (s.at(i) == 'A') s.at(ni++) = 0xc2;
	    else if (s.at(i) == 'a') s.at(ni++) = 0xe2;
	    else if (s.at(i) == 'I') s.at(ni++) = 0xce;
	    else if (s.at(i) == 'i') s.at(ni++) = 0xee;
	    else { s.at(ni++) = '^'; s.at(ni++) = s.at(i); }
	} else {
	    s.at(ni++) = s.at(i);
	}
	extrachar = ' ';
    }
    s.truncate(ni);
}


bool mainForm::dbOpenRead() {
  if ( !(dbFile = gdbm_open(dbFileName, 0, GDBM_READER, 0666, 0)) ) {
    qWarning("*** Warning: cannot open database for reading: %s", gdbm_strerror((gdbm_error) gdbm_errno));
    return false;
  }
  return true;
}

bool mainForm::dbOpenWrite() {
  if ( !(dbFile = gdbm_open (dbFileName, 0, GDBM_WRCREAT, 0666, 0)) ) {
      qWarning("*** Warning: cannot open database for writing: %s", gdbm_strerror((gdbm_error) gdbm_errno));
      return false;
  }
  return true;
}

void mainForm::dbClose() {
  if (dbFile) {
    gdbm_close(dbFile);
    dbFile = NULL;  
  }
}


// Called from the destructor.
void mainForm::destroy()
{
    if (pthread) {
      pthread->wait();
      delete pthread;
    }
    delete http;
    saveSettings();
}

